// A simple demonstration application showing the corrupted D2D windows
#include <stdio.h>

#include <windows.h>
#include <d2d1.h>
#include <dwrite.h>

const char appName[] = "KruptD2D";
const char className[] = "KruptD2DWindow";
const char kclassName[] = "ChildD2D";

struct ChildD2D {
	ID2D1HwndRenderTarget *pRenderTarget;
	HWND hwnd;

	ChildD2D();
	void EnsureRenderTarget();
	void DropRenderTarget();
	void Paint();
};

struct KruptD2D {
	HINSTANCE hInstance;
	HWND wMain;
	ChildD2D cd2d;
	ChildD2D cd2d2;

	KruptD2D();
};

static KruptD2D app;
static IDWriteFactory *pIDWriteFactory = 0;
static ID2D1Factory *pD2DFactory = 0;

static void LoadD2D() {
	static HMODULE hDLLD2D = 0;
	static HMODULE hDLLDWrite = 0;
	typedef HRESULT (WINAPI *D2D1CFSig)(D2D1_FACTORY_TYPE factoryType, REFIID riid,
		CONST D2D1_FACTORY_OPTIONS *pFactoryOptions, IUnknown **factory);
	typedef HRESULT (WINAPI *DWriteCFSig)(DWRITE_FACTORY_TYPE factoryType, REFIID iid,
		IUnknown **factory);

	hDLLD2D = ::LoadLibrary(TEXT("D2D1.DLL"));
	if (hDLLD2D) {
		D2D1CFSig fnD2DCF = (D2D1CFSig)::GetProcAddress(hDLLD2D, "D2D1CreateFactory");
		if (fnD2DCF) {
			// A single threaded factory as always draw on the GUI thread
			fnD2DCF(D2D1_FACTORY_TYPE_SINGLE_THREADED,
				__uuidof(ID2D1Factory),
				0,
				reinterpret_cast<IUnknown**>(&pD2DFactory));
		}
	}
	hDLLDWrite = ::LoadLibrary(TEXT("DWRITE.DLL"));
	if (hDLLDWrite) {
		DWriteCFSig fnDWCF = (DWriteCFSig)::GetProcAddress(hDLLDWrite, "DWriteCreateFactory");
		if (fnDWCF) {
			fnDWCF(DWRITE_FACTORY_TYPE_SHARED,
				__uuidof(IDWriteFactory),
				reinterpret_cast<IUnknown**>(&pIDWriteFactory));
		}
	}
}

KruptD2D::KruptD2D() {
	hInstance = 0;
	wMain = 0;
}

ChildD2D::ChildD2D() {
	pRenderTarget = 0;
	hwnd = 0;
}

void ChildD2D::EnsureRenderTarget() {
	if (pD2DFactory && !pRenderTarget) {
		RECT rc;
		GetClientRect(hwnd, &rc);

		D2D1_SIZE_U size = D2D1::SizeU(rc.right - rc.left, rc.bottom - rc.top);

		// Create a Direct2D render target.
		pD2DFactory->CreateHwndRenderTarget(
			D2D1::RenderTargetProperties(),
			D2D1::HwndRenderTargetProperties(hwnd, size),
			&pRenderTarget);
	}
}

void ChildD2D::DropRenderTarget() {
	if (pRenderTarget) {
		pRenderTarget->Release();
		pRenderTarget = 0;
	}
}


void ChildD2D::Paint() {
	//PAINTSTRUCT ps;
	//::BeginPaint(hwnd, &ps);
	EnsureRenderTarget();
	pRenderTarget->BeginDraw();

	RECT rc;
	GetClientRect(hwnd, &rc);

	D2D_COLOR_F colBack = {1.0, 1.0, 1.0, 1.0};
	ID2D1SolidColorBrush *pBrushBack = 0;
	D2D_COLOR_F colFore0 = {0.0, 0.5, 0.5, 1.0};
	ID2D1SolidColorBrush *pBrushFore0 = 0;
	D2D_COLOR_F colFore1 = {1.0, 0.8, 0.0, 1.0};
	ID2D1SolidColorBrush *pBrushFore1 = 0;

	HRESULT hr = pRenderTarget->CreateSolidColorBrush(colBack, &pBrushBack);
	if (SUCCEEDED(hr) && pBrushBack) {
		hr = pRenderTarget->CreateSolidColorBrush(colFore0, &pBrushFore0);
		if (SUCCEEDED(hr) && pBrushFore0) {
			hr = pRenderTarget->CreateSolidColorBrush(colFore1, &pBrushFore1);
			if (SUCCEEDED(hr) && pBrushFore1) {
				D2D1_RECT_F rectangleBack = D2D1::RectF(rc.left, rc.top, rc.right, rc.bottom);
				pRenderTarget->FillRectangle(&rectangleBack, pBrushBack);
				D2D1_RECT_F rectangleFore = D2D1::RectF(
					rc.left + 10.5, rc.top + 10.5, rc.right - 10.5, rc.bottom - 10.5);
				pRenderTarget->DrawRectangle(&rectangleFore, pBrushFore1);
				wchar_t wszText[] = L"Sample Text";
				IDWriteTextFormat *pTextFormat;
				hr = pIDWriteFactory->CreateTextFormat(L"Verdana", NULL,
					static_cast<DWRITE_FONT_WEIGHT>(400),
					DWRITE_FONT_STYLE_NORMAL,
					DWRITE_FONT_STRETCH_NORMAL, 16.0, L"en-us", &pTextFormat);
				if (pRenderTarget && pTextFormat && pBrushFore0) {
					IDWriteTextLayout *pTextLayout = 0;
					HRESULT hr = pIDWriteFactory->CreateTextLayout(wszText, wcslen(wszText),
						pTextFormat,
						rectangleFore.right - rectangleFore.left, 
						rectangleFore.bottom - rectangleFore.top, &pTextLayout);
					if (SUCCEEDED(hr)) {
						D2D1_POINT_2F origin = {rectangleFore.left, 10};
						pRenderTarget->DrawTextLayout(origin, pTextLayout, pBrushFore0,
							D2D1_DRAW_TEXT_OPTIONS_NONE);
					}
				}
				if (pTextFormat)
					pTextFormat->Release();
			}
		}
	}

	if (pBrushBack) {						
		pBrushBack->Release();
	}
	if (pBrushFore0) {						
		pBrushFore0->Release();
	}
	if (pBrushFore1) {						
		pBrushFore1->Release();
	}
	
	hr = pRenderTarget->EndDraw();
	if (hr == D2DERR_RECREATE_TARGET) {
		DropRenderTarget();
	}
	//::EndPaint(hwnd, &ps);
}

LRESULT CALLBACK KWndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
	ChildD2D *childThis = reinterpret_cast<ChildD2D *>((::GetWindowLongPtr(hWnd, 0)));
	// childThis will be zero if WM_CREATE not seen yet
	if (childThis == 0) {
		if (iMessage == WM_CREATE) {
			// Associate ChildD2D object with window
			CREATESTRUCT *pCreate = reinterpret_cast<CREATESTRUCT *>(lParam);
			::SetWindowLongPtr(hWnd, 0, reinterpret_cast<LONG_PTR>(pCreate->lpCreateParams));
			return 0;
		} else {
			return ::DefWindowProc(hWnd, iMessage, wParam, lParam);
		}
	}
	switch (iMessage) {
	case WM_NCDESTROY:
		::SetWindowLongPtr(hWnd, 0, 0);
		return ::DefWindowProc(hWnd, iMessage, wParam, lParam);
	case WM_PAINT:
		childThis->Paint();
		return 0;
	case WM_SIZE:
		childThis->DropRenderTarget();
		return DefWindowProc(hWnd, iMessage, wParam, lParam);
	default:
		return DefWindowProc(hWnd, iMessage, wParam, lParam);
	}
}

static void TurnOffHorizontalScroll(HWND hWnd) {
	SCROLLINFO sci = { sizeof(sci), 0, 0, 0, 0, 0, 0 };
	sci.fMask = SIF_PAGE | SIF_RANGE;
	::GetScrollInfo(hWnd, SB_HORZ, &sci);
	sci.nMin = 0;
	sci.nMax = 0;
	sci.nPage = 10;
	sci.nPos = 0;
	sci.nTrackPos = 1;
	::SetScrollInfo(hWnd, SB_HORZ, &sci, TRUE);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
	const DWORD styleFlags = WS_VISIBLE | WS_CHILD | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | 
		WS_VSCROLL | WS_HSCROLL | WS_CLIPSIBLINGS;
	switch (iMessage) {
	case WM_CREATE:
		app.cd2d.hwnd = ::CreateWindow(kclassName, "Source", styleFlags,
		              0, 0, 300, 300,
		              hWnd, 0, app.hInstance, &app.cd2d);
		app.cd2d2.hwnd = ::CreateWindow(kclassName, "Source2", styleFlags,
		              40, 40, 300, 300,
		              hWnd, 0, app.hInstance, &app.cd2d2);
		::SetFocus(app.cd2d.hwnd);

		TurnOffHorizontalScroll(app.cd2d.hwnd);
		return 0;

	case WM_CLOSE:
		::PostQuitMessage(0);
		return 0;

	default:
		return DefWindowProc(hWnd, iMessage, wParam, lParam);
	}
}

static void RegisterWindowClass() {
	const char resourceName[] = "KruptD2D";

	WNDCLASS wndclass;
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = app.hInstance;
	wndclass.hIcon = 0;
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = className;

	if (!::RegisterClass(&wndclass))
		::exit(FALSE);

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = KWndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = sizeof(ChildD2D *);
	wndclass.hInstance = app.hInstance;
	wndclass.hIcon = 0;
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = 0; // (HBRUSH)(COLOR_WINDOW+1);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = kclassName;

	if (!::RegisterClass(&wndclass))
		::exit(FALSE);
}

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR lpszCmdLine, int nCmdShow) {
	LoadD2D();

	app.hInstance = hInstance;

	RegisterWindowClass();

	app.wMain = ::CreateWindowEx(
	             WS_EX_CLIENTEDGE,
	             className,
	             "Demonstration",
	             WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
	             WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
	             WS_MAXIMIZE | WS_CLIPCHILDREN  | WS_CLIPSIBLINGS,
				 CW_USEDEFAULT, CW_USEDEFAULT, 
				 600, 800, 
	             NULL,
	             NULL,
	             app.hInstance,
	             0);

	::ShowWindow(app.wMain, nCmdShow);

	bool going = true;
	MSG msg;
	msg.wParam = 0;
	while (going) {
		going = GetMessage(&msg, NULL, 0, 0) != FALSE;
		if (going) {
			if (TranslateAccelerator(app.wMain, NULL, &msg) == 0) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}

	return msg.wParam;
}
